@@ -121,6 +121,64 @@ has _validator => sub { Swagger2::SchemaValidator->new; };
 
 =head1 HELPERS
 
+=head2 dispatch_to_swagger
+
+  $bool = $c->dispatch_to_swagger(\%data);
+
+This helper can be used to handle WebSocket requests with swagger.
+See L<Swagger2::Guides::WebSocket> for details.
+
+This helper is EXPERIMENTAL.
+
+=cut
+
+sub dispatch_to_swagger {
+  return undef unless $_[1]->{op} and $_[1]->{id} and ref $_[1]->{params} eq 'HASH';
+
+  my ($c, $data) = @_;
+  my $self     = $c->stash('swagger.plugin');
+  my $reply    = sub { $_[0]->send({json => {code => $_[2] || 200, id => $data->{id}, body => $_[1]}}) };
+  my $op_info  = $self->{op_info}{$data->{op}} or return $c->$reply(_error('Unknown operationId.'), 400);
+  my $sc_class = $op_info->{class} ||= _find_controller($c, $op_info->{controller});
+  my ($input, $method_ref, @errors);
+
+  return $c->$reply(_error('Controller could not be loaded.'), 501) unless $sc_class;
+  return $c->$reply(_error(qq(Method "$op_info->{method}" not implemented.)), 501)
+    unless $method_ref = $sc_class->can($op_info->{method});
+
+  for my $p (@{$op_info->{spec}{parameters} || []}) {
+    my $name  = $p->{name};
+    my $value = $data->{params}{$name} // $p->{default};
+    my @e     = $self->_validate_value($p, $name => $value);
+    $input->{$name} = $value unless @e;
+    push @errors, @e;
+  }
+
+  return $c->$reply({errors => \@errors}, 400) if @errors;
+  return Mojo::IOLoop->delay(
+    sub {
+      my $delay = shift;
+      my $sc = $delay->data->{sc} = $sc_class->new(tx => $c->tx);
+      $sc->stash(swagger_operation_spec => $op_info->{spec});
+      $sc->$method_ref($input, $delay->begin);
+    },
+    sub {
+      my $delay  = shift;
+      my $data   = shift;
+      my $status = shift || 200;
+      my $format = $op_info->{spec}{responses}{$status} || $op_info->{spec}{responses}{default} || undef;
+      my @errors
+        = !$format ? $self->_validator->validate($data, {})
+        : $format->{schema} ? $self->_validator->validate($data, $format->{schema})
+        :                     ();
+
+      return $c->$reply($data, $status) unless @errors;
+      warn "[Swagger2] Invalid response: @errors\n" if DEBUG;
+      $c->$reply({errors => \@errors}, 500);
+    },
+  );
+}
+
 =head2 render_swagger
 
   $c->render_swagger(\%err, \%data, $status);
@@ -157,6 +215,19 @@ C<%config> can contain these parameters:
 
 This argument will be passed on to L<JSON::Validator/coerce>.
 
+=item * ensure_swagger_response
+
+  $app->plugin(swagger2 => {
+    ensure_swagger_response => {
+      exception => "Internal server error.",
+      not_found => "Not found.",
+    }
+  });
+
+Adds a L<before_render|Mojolicious/HOOKS> hook which will make sure
+"exception" and "not_found" responses are in JSON format. There is no need to
+specify "exception" and "not_found" if you are happy with the defaults.
+
 =item * route
 
 Need to hold a Mojolicious route object. See L</Protected API> for an example.
@@ -213,7 +284,8 @@ sub register {
   local $config->{coerce} = $config->{coerce} || $ENV{SWAGGER_COERCE_VALUES};
   $self->_validator->coerce($config->{coerce}) if $config->{coerce};
   $self->url($swagger->url);
-  $app->helper(render_swagger => \&render_swagger) unless $app->renderer->get_helper('render_swagger');
+  $app->helper(dispatch_to_swagger => \&dispatch_to_swagger) unless $app->renderer->get_helper('dispatch_to_swagger');
+  $app->helper(render_swagger      => \&render_swagger)      unless $app->renderer->get_helper('render_swagger');
 
   $r = $config->{route};
 
@@ -225,6 +297,9 @@ sub register {
     $r = $app->routes->any($swagger->base_url->path->to_string);
     $r->to(swagger => $swagger);
   }
+  if (my $ws = $config->{ws}) {
+    $ws->to('swagger.plugin' => $self);
+  }
 
   for my $path (sort { length $a <=> length $b } keys %$paths) {
     $paths->{$path}{'x-mojo-around-action'} ||= $swagger->api_spec->get('/x-mojo-around-action');
@@ -248,22 +323,44 @@ sub register {
       warn "[Swagger2] Add route $http_method $route_path\n" if DEBUG;
     }
   }
-}
 
-sub _find_controller {
-  my ($self, $c, $moniker) = @_;
-  my $controller = $moniker;
-
-  return $controller if $controller =~ /::/ and !defined load_class $controller;
-  $controller =~ s!s$!!;    # plural to singular, "::Pets" to "::Pet"
-
-  for my $ns (@{$c->app->routes->namespaces}) {
-    my $class = "${ns}::$controller";
-    return $class unless defined load_class $class;
+  # EXPERIMENTAL: Need documentation and probably a better route name()
+  if (my $title = $swagger->api_spec->get('/info/title')) {
+    my $md5
+      = Mojo::Util::md5_sum(
+      Data::Dumper->new([$swagger->api_spec->data])->Indent(0)->Pair('=>')->Purity(1)->Quotekeys(1)->Sortkeys(1)
+        ->Terse(1)->Useqq(1)->Dump);
+    $title =~ s!\W!_!g;
+    $r->get("/$md5", [format => [qw( json )]], {format => 'json'})
+      ->to(cb => sub { shift->render(json => $swagger->api_spec->data) })->name(lc $title);
   }
 
-  $c->app->log->error(qq(Could not find controller class for "$moniker": $@));
-  return;
+  $swagger->api_spec->data->{basePath} = $r->to_string;
+
+  if ($config->{ensure_swagger_response}) {
+    $self->_ensure_swagger_response($app, $config->{ensure_swagger_response}, $swagger);
+  }
+}
+
+sub _ensure_swagger_response {
+  my ($self, $app, $responses, $swagger) = @_;
+  my $base_path = $swagger->api_spec->data->{basePath};
+
+  $responses->{exception} ||= 'Internal server error.';
+  $responses->{not_found} ||= 'Not found.';
+  $base_path = qr{^$base_path};
+
+  $app->hook(
+    before_render => sub {
+      my ($c, $args) = @_;
+
+      return unless my $template = $args->{template};
+      return unless my $msg      = $responses->{$template};
+      return unless $c->req->url->path->to_string =~ $base_path;
+
+      $args->{json} = _error($msg);
+    }
+  );
 }
 
 sub _find_controller_and_method {
@@ -283,34 +380,34 @@ sub _find_controller_and_method {
 sub _generate_request_handler {
   my ($self,       $op_spec) = @_;
   my ($controller, $method)  = $self->_find_controller_and_method($op_spec);
-  my ($controller_class, $handler);
   my $defaults = {swagger_operation_spec => $op_spec};
+  my $op_info = {controller => $controller, method => $method, spec => $op_spec};
 
-  $handler = sub {
+  my $handler = sub {
     my $c = shift;
     my ($method_ref, $v, $input);
 
-    unless ($controller_class ||= $self->_find_controller($c, $controller)) {
-      return $c->render_swagger($self->_not_implemented('Controller could not be loaded.'), {}, 501);
+    unless ($op_info->{class} ||= _find_controller($c, $controller)) {
+      return $c->render_swagger(_error('Controller could not be loaded.'), {}, 501);
     }
-    unless ($method_ref = $controller_class->can($method)) {
-      $method_ref = $controller_class->can(sprintf '%s_%s', $method, lc $c->req->method)
+    unless ($method_ref = $op_info->{class}->can($method)) {
+      $method_ref = $op_info->{class}->can(sprintf '%s_%s', $method, lc $c->req->method)
         and warn "HTTP method name is not used in method name lookup anymore!";
     }
     unless ($method_ref) {
       $c->app->log->error(
-        qq(Can't locate object method "$method" via package "$controller_class. (Something is wrong in @{[$self->url]})")
+        qq(Can't locate object method "$method" via package "$op_info->{class}". (Something is wrong in @{[$self->url]})")
       );
-      return $c->render_swagger($self->_not_implemented(qq(Method "$method" not implemented.)), {}, 501);
+      return $c->render_swagger(_error(qq(Method "$method" not implemented.)), {}, 501);
     }
 
-    bless $c, $controller_class;    # ugly hack?
+    bless $c, $op_info->{class};    # ugly hack?
     ($v, $input) = $self->_validate_input($c, $op_spec);
 
     return $c->render_swagger($v, {}, 400) if @{$v->{errors}};
     return $c->delay(
       sub {
-        $c->app->log->debug("Swagger2 calling $controller_class\->$method(\$input, \$cb)");
+        $c->app->log->debug("Swagger2 calling $op_info->{class}\->$method(\$input, \$cb)");
         $c->$method_ref($input, shift->begin);
       },
       sub {
@@ -318,13 +415,14 @@ sub _generate_request_handler {
         my $data   = shift;
         my $status = shift || 200;
         my $format = $op_spec->{responses}{$status} || $op_spec->{responses}{default} || undef;
-        my @err
+        my @errors
           = !$format ? $self->_validator->validate($data, {})
           : $format->{schema} ? $self->_validator->validate($data, $format->{schema})
           :                     ();
 
-        return $c->render_swagger({errors => \@err}, $data, 500) if @err;
-        return $c->render_swagger({}, $data, $status);
+        return $c->render_swagger({}, $data, $status) unless @errors;
+        warn "[Swagger2] Invalid response: @errors\n" if DEBUG;
+        $c->render_swagger({errors => \@errors}, $data, 500);
       },
     );
   };
@@ -342,14 +440,10 @@ sub _generate_request_handler {
     };
   }
 
+  $self->{op_info}{$op_spec->{operationId}} = $op_info;
   return $defaults, $handler;
 }
 
-sub _not_implemented {
-  my ($self, $message) = @_;
-  return {errors => [{message => $message, path => '/'}]};
-}
-
 sub _on_route_added {
   my ($self, $r) = @_;
   my $op_spec    = $r->pattern->defaults->{swagger_operation_spec};
@@ -372,10 +466,8 @@ sub _validate_input {
   my (%input, @errors);
 
   for my $p (@{$op_spec->{parameters} || []}) {
-    my ($in, $name) = @$p{qw( in name )};
-    my ($value, @e);
-
-    $value
+    my ($in, $name, $type) = @$p{qw( in name type )};
+    my $value
       = $in eq 'query'    ? $query->param($name)
       : $in eq 'path'     ? $c->stash($name)
       : $in eq 'header'   ? $headers->header($name)
@@ -383,36 +475,19 @@ sub _validate_input {
       : $in eq 'formData' ? $body->param($name)
       :                     "Invalid 'in' for parameter $name in schema definition";
 
-    $value //= $p->{default};
-
-    if (defined $value or Swagger2::_is_true($p->{required})) {
-      my $type = $p->{type} || 'object';
-      $value += 0 if $type =~ /^(?:integer|number)/ and $value =~ /^-?\d/;
-      $value = ($value eq 'false' or !$value) ? Mojo::JSON->false : Mojo::JSON->true if $type eq 'boolean';
-
-      my $schema = {properties => {$name => $p}, required => [$p->{required} ? ($p->{name}) : ()],};
-
-      # ugly hack
-      if (ref $p->{items} eq 'HASH' and $p->{items}{collectionFormat}) {
+    if ($type and defined($value //= $p->{default})) {
+      if (($type eq 'integer' or $type eq 'number') and $value =~ /^-?\d/) {
+        $value += 0;
+      }
+      elsif ($type eq 'boolean') {
+        $value = (!$value or $value eq 'false') ? Mojo::JSON->false : Mojo::JSON->true;
+      }
+      elsif (ref $p->{items} eq 'HASH' and $p->{items}{collectionFormat}) {
         $value = _coerce_by_collection_format($p->{items}, $value);
       }
-
-      if ($in eq 'body') {
-        warn "[Swagger2] Validate $in @{[$c->req->body]}\n" if DEBUG;
-        push @e,
-          map { $_->{path} = $_->{path} eq "/" ? "/$name" : "/$name$_->{path}"; $_; }
-          $self->_validator->validate($value, $p->{schema});
-      }
-      elsif (defined $value) {
-        warn "[Swagger2] Validate $in $name=$value\n" if DEBUG;
-        push @e, $self->_validator->validate({$name => $value}, $schema);
-      }
-      else {
-        warn "[Swagger2] Validate $in $name=undef()\n" if DEBUG;
-        push @e, $self->_validator->validate({}, $schema);
-      }
     }
 
+    my @e = $self->_validate_value($p, $name => $value);
     $input{$name} = $value unless @e;
     push @errors, @e;
   }
@@ -420,6 +495,34 @@ sub _validate_input {
   return {errors => \@errors}, \%input;
 }
 
+sub _validate_value {
+  my ($self, $p, $name, $value) = @_;
+  my $type = $p->{type} || 'object';
+  my @e;
+
+  return if !defined $value and !Swagger2::_is_true($p->{required});
+
+  my $schema = {properties => {$name => $p->{schema} || $p}, required => [$p->{required} ? ($name) : ()]};
+  my $in = $p->{in};
+
+  # ugly hack
+
+  if ($in eq 'body') {
+    warn "[Swagger2] Validate $in $name\n" if DEBUG;
+    return $self->_validator->validate({$name => $value}, $schema);
+  }
+  elsif (defined $value) {
+    warn "[Swagger2] Validate $in $name=$value\n" if DEBUG;
+    return $self->_validator->validate({$name => $value}, $schema);
+  }
+  else {
+    warn "[Swagger2] Validate $in $name=undef\n" if DEBUG;
+    return $self->_validator->validate({$name => $value}, $schema);
+  }
+
+  return;
+}
+
 # copy/paste from JSON::Validator
 sub _coerce_by_collection_format {
   my ($schema, $data) = @_;
@@ -435,6 +538,26 @@ sub _die {
   die "$_[1]: ", Mojo::Util::dumper($_[0]);
 }
 
+sub _error {
+  return {errors => [{message => $_[0], path => '/'}]};
+}
+
+sub _find_controller {
+  my ($c, $moniker) = @_;
+  my $controller = $moniker;
+
+  return $controller if $controller =~ /::/ and !defined load_class $controller;
+  $controller =~ s!s$!!;    # plural to singular, "::Pets" to "::Pet"
+
+  for my $ns (@{$c->app->routes->namespaces}) {
+    my $class = "${ns}::$controller";
+    return $class unless defined load_class $class;
+  }
+
+  $c->app->log->error(qq(Could not find controller class for "$moniker": $@));
+  return;
+}
+
 =head1 COPYRIGHT AND LICENSE
 
 Copyright (C) 2014-2015, Jan Henning Thorsen
